/*!@file
** @brief   The implementation of the facade for the clipboard.
** @remarks The facade simplifies the work with the clipboard.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
**
** @copyright (C) 2014-2019 Ueta System Design Studio\n
** This file is licensed under the terms of the MIT License.\n
** See <https://opensource.org/licenses/MIT>.
*/
#pragma once

/*!
** @brief   The facade for the clipboard.
** @remarks -
*/
class CClipboard
{
public:

	/*!
	** @brief   Allocates a new global memory object and copies the specified string into it.
	** @param   [in] psz the string to copy.
	** @returns a handle to the newly allocated memory object if successful, or NULL otherwise.
	** @remarks -
	*/
	static HGLOBAL GlobalAllocString(const wchar_t* psz)
	{
		size_t cch = wcslen(psz) + 1; // including the terminating null character

		if (HGLOBAL hGlobal = ::GlobalAlloc(GMEM_MOVEABLE, sizeof(wchar_t) * cch))
		{
			if (wchar_t* pszGlobal = static_cast<wchar_t*>(::GlobalLock(hGlobal)))
			{
				ATLVERIFY(wcscpy_s(pszGlobal, cch, psz) == 0);

				ATLVERIFY(::GlobalUnlock(hGlobal) == 0);
				ATLASSERT(::GetLastError() == NO_ERROR);

				return hGlobal;
			}
			ATLVERIFY(::GlobalFree(hGlobal) == NULL);
		}
		return NULL;
	}

	/*!
	** @brief   Deallocates a string allocated in the global memory.
	** @param   [in] hGlobal a handle to the global memory object.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	static BOOL GlobalFreeString(HGLOBAL hGlobal)
	{
		if (::GlobalFree(hGlobal) == NULL)
		{
			return TRUE;
		}
		return FALSE;
	}

	/*!
	** @brief   Clears the clipboard and then places the specified data.
	** @param   [in] uFormat the clipboard format.
	** @param   [in] hMem a handle to the data in the specified format.
	** @param   [in] hWndOwner a handle to the window to be associated with the clipboard.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	static BOOL SetData(UINT uFormat, HANDLE hMem, HWND hWndOwner = ::GetDesktopWindow())
	{
		BOOL bSuccess = FALSE;

		if (::OpenClipboard(hWndOwner))
		{
			if (::EmptyClipboard())
			{
				if (::SetClipboardData(uFormat, hMem) != NULL)
				{
					bSuccess = TRUE;
				}
			}
			ATLVERIFY(::CloseClipboard());
		}
		return bSuccess;
	}

	/*!
	** @brief   Clears the clipboard and then places the specified text data.
	** @param   [in] pszText a null-terminated string to set to the clipboard.
	** @param   [in] hWndOwner a handle to the window to be associated with the clipboard.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	static BOOL SetText(const wchar_t* pszText, HWND hWndOwner = ::GetDesktopWindow())
	{
		if (HGLOBAL hGlobal = GlobalAllocString(pszText))
		{
			if (SetData(CF_UNICODETEXT, hGlobal, hWndOwner)) // the system owns the data
			{
				return TRUE;
			}
			ATLVERIFY(GlobalFreeString(hGlobal));
		}
		return FALSE;
	}
};
