/*!@file
** @brief   The implementation of the DWM-aware functions.
** @remarks More features for the Desktop Window Manager APIs are provided.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
**
** @copyright (C) 2014-2019 Ueta System Design Studio\n
** This file is licensed under the terms of the MIT License.\n
** See <https://opensource.org/licenses/MIT>.
*/
#pragma once

#pragma comment(lib, "dwmapi.lib")

#include <dwmapi.h>

/*!
** @brief   Retrieves the extended frame bounds of the specified window.
** @param   [in] hWnd a handle to the window.
** @param   [out] lpRect a pointer to a RECT structure to receive the bounds.
** @returns TRUE if successful, or FALSE otherwise.
** @remarks -
** @note    The extended frame bounds of a window can be inside the bounding
**          rectangle of the window. Windows 10 has a drop shadow affect around
**          three edges of a window, and it seems that is classed as being part
**          of the window.
*/
BOOL DwmGetExtendedFrameBounds(HWND hWnd, LPRECT lpRect)
{
	HRESULT hResult = ::DwmGetWindowAttribute(hWnd, DWMWA_EXTENDED_FRAME_BOUNDS, lpRect, sizeof(*lpRect));

	return SUCCEEDED(hResult) ? TRUE : FALSE;
}

/*!
** @brief   Retrieves information about the specified window.
** @param   [in] hWnd a handle to the window whose information is to be retrieved.
** @param   [out] pwi a pointer to a WINDOWINFO structure to receive the information.
** @returns TRUE if successful, or FALSE otherwise.
** @remarks -
*/
BOOL DwmGetWindowInfo(HWND hWnd, PWINDOWINFO pwi)
{
	if (::GetWindowInfo(hWnd, pwi))
	{
		#if 1 // extended frame bounds aware
		{
			CRect& rcWindow = static_cast<CRect&>(pwi->rcWindow);

			CRect rcExtendedFrameBounds;

			if (DwmGetExtendedFrameBounds(hWnd, rcExtendedFrameBounds))
			{
				// ATLASSERT(rcExtendedFrameBounds.Width()  >= rcWindow.Width());
				// ATLASSERT(rcExtendedFrameBounds.Height() >= rcWindow.Height());

				rcWindow = rcExtendedFrameBounds;
			}
		}
		#endif

		return TRUE;
	}
	return FALSE;
}

/*!
** @brief   Checks if the specified window is cloaked.
** @param   [in] hWnd a handle to the window to be tested.
** @returns TRUE if cloaked, or FALSE otherwise.
** @remarks -
*/
BOOL DwmIsWindowCloaked(HWND hWnd)
{
	int dwCloaked = 0;

	HRESULT hResult = ::DwmGetWindowAttribute(hWnd, DWMWA_CLOAKED, &dwCloaked, sizeof(dwCloaked));

	return SUCCEEDED(hResult) && dwCloaked ? TRUE : FALSE;
}

/*!
** @brief   Checks if the specified window is visible.
** @param   [in] hWnd a handle to the window to be tested.
** @returns TRUE if visible, or FALSE otherwise.
** @remarks -
*/
BOOL DwmIsWindowVisible(HWND hWnd)
{
	if (!::IsWindowVisible(hWnd))
	{
		return FALSE;
	}
	if (DwmIsWindowCloaked(hWnd))
	{
		return FALSE;
	}
	return TRUE;
}

/*!
** @brief   Changes the size, position, and Z-order of a window.
** @param   [in] hWnd a handle to the window.
** @param   [in] hWndInsertAfter a handle to the window to precede in the Z-order.
** @param   [in] x the new position of the left side of the window.
** @param   [in] y the new position of the top of the window.
** @param   [in] cx the new width of the window.
** @param   [in] cy the new height of the window.
** @param   [in] uFlags the sizing and positioning flags.
** @returns TRUE if successful, or FALSE otherwise.
** @remarks -
*/
BOOL DwmSetWindowPos(HWND hWnd, HWND hWndInsertAfter, int x, int y, int cx, int cy, UINT uFlags)
{
	#if 1 // extended frame bounds aware
	{
		CRect rcWindow;

		if (::GetWindowRect(hWnd, rcWindow))
		{
			CRect rcExtendedFrameBounds;

			if (DwmGetExtendedFrameBounds(hWnd, rcExtendedFrameBounds))
			{
				// ATLASSERT(rcWindow.left >= rcExtendedFrameBounds.left);
				// ATLASSERT(rcWindow.top  >= rcExtendedFrameBounds.top);

				x += rcWindow.left - rcExtendedFrameBounds.left;
				y += rcWindow.top  - rcExtendedFrameBounds.top;

				// ATLASSERT(rcExtendedFrameBounds.Width()  >= rcWindow.Width());
				// ATLASSERT(rcExtendedFrameBounds.Height() >= rcWindow.Height());

				int dx = rcExtendedFrameBounds.Width()  - rcWindow.Width();
				int dy = rcExtendedFrameBounds.Height() - rcWindow.Height();

				cx -= min(cx, dx);
				cy -= min(cy, dy);
			}
		}
	}
	#endif

	if (::SetWindowPos(hWnd, hWndInsertAfter, x, y, cx, cy, uFlags))
	{
		;

		return TRUE;
	}
	return FALSE;
}
