/*!@file
** @brief   The implementation of the main dialog box.
** @remarks This utility is a simple dialog-based application.
** @author  UETA, Shin-ichi <ueta@usdesign.jp>
**
** @copyright (C) 2014-2019 Ueta System Design Studio\n
** This file is licensed under the terms of the MIT License.\n
** See <https://opensource.org/licenses/MIT>.
*/
#pragma once

#include "about.h"
#include "browse.h"
#include "datetime.h"
#include "file.h"
#include "options.h"

template <class T>
class CPropertyPageEx;

/*!
** @brief   The customization of the property sheet.
** @remarks -
*/
class ATL_NO_VTABLE CPropertySheetEx :
		public CPropertySheetImpl<CPropertySheetEx>,
		public CAboutBoxSiteImpl<CPropertySheetEx>,
		public CMessageFilter
{
public:

	/*!
	** @brief   Constructs a new instance of the CPropertySheetEx class with the default value.
	** @remarks -
	*/
	CPropertySheetEx() : CPropertySheetImpl<CPropertySheetEx>(IDR_APP)
	{
		m_psh.dwFlags |= PSH_NOCONTEXTHELP;
	}

	/*!
	** @brief   Displays an alert box with a message about the specified error.
	** @param   [in] hResult a standard HRESULT value that indicates an error.
	** @param   [in] pszDescription the brief description of the error.
	** @returns No return value.
	** @remarks -
	*/
	void Alert(HRESULT hResult, LPCTSTR pszDescription = NULL)
	{
		CString message = AtlGetErrorDescription(hResult);

		if (pszDescription)
		{
			(message += TEXT('\n')) += pszDescription;
		}
		AtlMessageBox(m_hWnd, LPCTSTR(message), IDR_APP, MB_ICONERROR);
	}

	/*!
	** @brief   Associates a large or small icon with the window.
	** @param   [in] icon specifies the icon resource to be loaded.
	** @param   [in] bBigIcon the type of the icon.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL SetIcon(_U_STRINGorID icon, BOOL bBigIcon = TRUE)
	{
		int cxIcon = ::GetSystemMetrics(bBigIcon ? SM_CXICON : SM_CXSMICON);
		int cyIcon = ::GetSystemMetrics(bBigIcon ? SM_CYICON : SM_CYSMICON);

		if (HICON hIcon = AtlLoadIconImage(icon, LR_DEFAULTCOLOR, cxIcon, cyIcon))
		{
			CWindow::SetIcon(hIcon, bBigIcon);

			return TRUE;
		}
		return FALSE;
	}

	/*!
	** @brief   Adds the supplied page with the rightmost tab in the property sheet.
	** @param   [in] page a reference to the page to be added to the property sheet.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	template <class T>
	BOOL AddPage(CPropertyPageEx<T>& page)
	{
		if (CPropertySheetImpl<CPropertySheetEx>::AddPage(page))
		{
			page.m_pPropSheet = this;

			return TRUE;
		}
		return FALSE;
	}

	BEGIN_MSG_MAP(CPropertySheetEx)
		MSG_WM_DESTROY(OnDestroy)
		CHAIN_MSG_MAP(CPropertySheetImpl<CPropertySheetEx>)
		CHAIN_MSG_MAP(CAboutBoxSiteImpl<CPropertySheetEx>)
	END_MSG_MAP()

	/*!
	** @brief   Filters window messages before they are dispatched to windows.
	** @param   [in] pMsg a pointer to an MSG structure that contains a message.
	** @returns TRUE if the message should not be dispatched, or FALSE otherwise.
	** @remarks -
	*/
	virtual BOOL PreTranslateMessage(MSG* pMsg)
	{
		if (!IsDialogMessage(pMsg))
		{
			return FALSE;
		}
		return TRUE;
	}

	/*!
	** @brief   Called when the property sheet is being initialized.
	** @returns No return value.
	** @remarks -
	*/
	void OnSheetInitialized()
	{
		ATLVERIFY(ModifyStyleEx(0, WS_EX_APPWINDOW));

		ATLVERIFY(SetIcon(IDR_APP, TRUE));
		ATLVERIFY(SetIcon(IDR_APP, FALSE));

		ATLVERIFY(BindAboutBox());

		CMessageLoop* pMessageLoop = _Module.GetMessageLoop();
		ATLASSERT(pMessageLoop);
		ATLVERIFY(pMessageLoop->AddMessageFilter(this));
	}

	/*!
	** @brief   Called when the property sheet is about to be destroyed.
	** @returns No return value.
	** @remarks -
	*/
	void OnDestroy()
	{
		CMessageLoop* pMessageLoop = _Module.GetMessageLoop();
		ATLASSERT(pMessageLoop);
		ATLVERIFY(pMessageLoop->RemoveMessageFilter(this));

		int nExitCode = GetResult();

		::PostQuitMessage(nExitCode);
	}

	/*!
	** @brief   Called when the user has clicked the Apply button.
	** @returns One of the return values for the PSN_APPLY notification.
	** @remarks -
	*/
	virtual int OnApply() = 0;
};

/*!
** @brief   The base class of individual pages of the property sheet.
** @remarks -
*/
template <class T>
class ATL_NO_VTABLE CPropertyPageEx :
		public CPropertyPageImpl<T>,
		public CWinDataExchange<T>,
		public CIdleHandler
{
private:

	friend class CPropertySheetEx;

	CPropertySheetEx* m_pPropSheet;

public:

	/*!
	** @brief   Initializes a new CPropertyPageEx object.
	** @remarks -
	*/
	CPropertyPageEx()
	{
		m_pPropSheet = NULL;
	}

	/*!
	** @brief   Retrieves the property sheet that contains this page.
	** @returns A reference to the CPropertySheetEx object.
	** @remarks -
	*/
	CPropertySheetEx& GetPropertySheet() const
	{
		ATLASSERT(m_pPropSheet);

		return *m_pPropSheet;
	}

	/*!
	** @brief   Returns the zero-based index of this page in the property sheet.
	** @returns The zero-based index of this page if successful, or -1 otherwise.
	** @remarks -
	*/
	int GetIndex() const
	{
		return GetPropertySheet().IdToIndex(T::IDD);
	}

	/*!
	** @brief   Checks if this page is the last page in the property sheet.
	** @returns true if this page is the last page, or false otherwise.
	** @remarks -
	*/
	bool IsLastPage() const
	{
		int nPageCount = GetPropertySheet().GetPageCount();

		return nPageCount > 0 ? 1 + GetIndex() == nPageCount : false;
	}

	/*!
	** @brief   Displays an alert box with a message about the specified error.
	** @param   [in] hResult a standard HRESULT value that indicates an error.
	** @param   [in] pszDescription the brief description of the error.
	** @returns No return value.
	** @remarks -
	*/
	void Alert(HRESULT hResult, LPCTSTR pszDescription = NULL)
	{
		GetPropertySheet().Alert(hResult, pszDescription);
	}

	/*!
	** @brief   Enables or disables the specified control in the dialog box.
	** @param   [in] nCtrlID the identifier of the control to be enabled or disabled.
	** @param   [in] bEnable specifies whether the control is to be enabled or disabled.
	** @returns TRUE if the control was previously disabled, or FALSE otherwise.
	** @remarks -
	*/
	BOOL EnableDlgCtrl(int nCtrlID, BOOL bEnable = TRUE)
	{
		T* pT = static_cast<T*>(this);
		CWindow wndCtrl = pT->GetDlgItem(nCtrlID);
		ATLASSERT(wndCtrl);

		return wndCtrl.EnableWindow(bEnable);
	}

	/*!
	** @brief   Hides the specified control in the dialog box.
	** @param   [in] nCtrlID the identifier of the control to be hidden.
	** @returns TRUE if the control was previously visible, or FALSE otherwise.
	** @remarks -
	*/
	BOOL HideDlgCtrl(int nCtrlID)
	{
		T* pT = static_cast<T*>(this);
		CWindow wndCtrl = pT->GetDlgItem(nCtrlID);
		ATLASSERT(wndCtrl);

		return wndCtrl.ShowWindow(SW_HIDE);
	}

	/*!
	** @brief   Synchronizes the value of a data member with a control in the dialog box.
	** @param   [in] nID the identifier of a control in the dialog box.
	** @param   [in] hIcon a reference to a data member in the dialog box.
	** @param   [in] bSave indicates the direction of the data transfer.
	** @returns No return value.
	** @remarks -
	*/
	void DDX_Icon(UINT nID, HICON& hIcon, BOOL bSave)
	{
		T* pT = static_cast<T*>(this);
		HWND hWndCtrl = pT->GetDlgItem(nID);
		ATLASSERT(hWndCtrl);

		if (bSave)
			hIcon = CStatic(hWndCtrl).GetIcon();
		else
			CStatic(hWndCtrl).SetIcon(hIcon);
	}

	/*!
	** @brief   Synchronizes the value of a data member with a control in the dialog box.
	** @param   [in] nID the identifier of a control in the dialog box.
	** @param   [in] nVal a reference to a data member in the dialog box.
	** @param   [in] bSave indicates the direction of the data transfer.
	** @returns No return value.
	** @remarks -
	*/
	void DDX_Option(UINT nID, INT& nVal, BOOL bSave)
	{
		DDX_Index<CComboBox>(nID, nVal, bSave);
	}

	/*!
	** @brief   Synchronizes the value of a data member with a control in the dialog box.
	** @param   [in] nID the identifier of a control in the dialog box.
	** @param   [in] ftVal a reference to a data member in the dialog box.
	** @param   [in] bSave indicates the direction of the data transfer.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL DDX_DateTime(UINT nID, FILETIME& ftVal, BOOL bSave)
	{
		T* pT = static_cast<T*>(this);
		HWND hWndCtrl = pT->GetDlgItem(nID);
		ATLASSERT(hWndCtrl);
		BOOL bSuccess = TRUE;

		if (bSave)
			bSuccess = CDateTimePickerCtrlEx(hWndCtrl).GetFileTime(&ftVal) != GDT_ERROR;
		else
			bSuccess = CDateTimePickerCtrlEx(hWndCtrl).SetFileTime(GDT_VALID, &ftVal);

		if (!bSuccess)
		{
			pT->OnDataExchangeError(nID, bSave);
		}
		else if (bSave) // validation
		{
			;
		}
		return bSuccess;
	}

	BEGIN_MSG_MAP(CPropertyPageEx)
		MSG_WM_INITDIALOG(OnInitDialog)
		MSG_WM_DESTROY(OnDestroy)
		CHAIN_MSG_MAP(CPropertyPageImpl<T>)
	END_MSG_MAP()

	/*!
	** @brief   Called in response to the WM_INITDIALOG message.
	** @param   [in] hWndFocus a handle to the control to receive the default input focus.
	** @param   [in] lInitParam additional initialization data.
	** @returns TRUE to set the default input focus, or FALSE otherwise.
	** @remarks -
	*/
	virtual BOOL OnInitDialog(HWND hWndFocus, LPARAM lInitParam)
	{
		UNREFERENCED_PARAMETER(hWndFocus);
		UNREFERENCED_PARAMETER(lInitParam);

		ATLVERIFY(ExecuteDlgInit(T::IDD));

		CMessageLoop* pMessageLoop = _Module.GetMessageLoop();
		ATLASSERT(pMessageLoop);
		ATLVERIFY(pMessageLoop->AddIdleHandler(this));

		return TRUE;
	}

	/*!
	** @brief   Called when the page is about to be destroyed.
	** @returns No return value.
	** @remarks -
	*/
	virtual void OnDestroy()
	{
		CMessageLoop* pMessageLoop = _Module.GetMessageLoop();
		ATLASSERT(pMessageLoop);
		ATLVERIFY(pMessageLoop->RemoveIdleHandler(this));
	}

	/*!
	** @brief   Called when any control in the property page is changed.
	** @param   [in] uNotifyCode the notification code.
	** @param   [in] nID the identifier of the control sending the message.
	** @param   [in] hWndCtrl a handle to the control sending the message.
	** @returns No return value.
	** @remarks -
	*/
	void OnChange(UINT uNotifyCode, int nID, HWND hWndCtrl)
	{
		UNREFERENCED_PARAMETER(uNotifyCode);
		UNREFERENCED_PARAMETER(nID);
		UNREFERENCED_PARAMETER(hWndCtrl);

		SetModified();
		SetMsgHandled(FALSE);
	}

	/*!
	** @brief   Called when any control in the property page is changed.
	** @param   [in] pnmh a pointer to information about the notification message.
	** @returns The result of the message processing.
	** @remarks -
	*/
	LRESULT OnChange(LPNMHDR pnmh)
	{
		UNREFERENCED_PARAMETER(pnmh);

		SetModified();
		SetMsgHandled(FALSE);

		return LRESULT(0);
	}

	/*!
	** @brief   Called when the user has clicked the Apply button.
	** @returns One of the return values for the PSN_APPLY notification.
	** @remarks -
	*/
	int OnApply()
	{
		T* pT = static_cast<T*>(this);

		int iResult = PSNRET_INVALID_NOCHANGEPAGE;

		if (pT->DoDataExchange(DDX_SAVE))
		{
			iResult = IsLastPage() ? GetPropertySheet().OnApply() : PSNRET_NOERROR;
		}
		return iResult;
	}
};

/*!
** @brief   The implementation of the General property page.
** @remarks -
*/
class CGeneralPage :
		public CPropertyPageEx<CGeneralPage>,
		public CGeneralOptions<CGeneralPage>
{
public:

	enum {IDD = IDD_GENERAL}; // specifies the dialog template resource ID.

public:

	typedef CGeneralOptions<CGeneralPage> CGeneralOptions; ///< A synonym for the template class.
	typedef CGeneralContext<CGeneralPage> CGeneralContext; ///< A synonym for the template class.

	typedef CGeneralOptions CHandlingOptions; ///< The abstract name of the CGeneralOptions class.
	typedef CGeneralContext CHandlingContext; ///< The abstract name of the CGeneralContext class.

private:

	friend class CGeneralOptions;
	friend class CDialog;

	CDateTimePickerCtrlExImpl m_wndDateTimePickers[3];

	HICON m_hIcon;

	CString m_strDisplayName;
	CString m_strTypeName;
	CString m_strLocation;

	bool m_bDirectory;
	bool m_bRecursive;

	bool m_bExcludeSystem;
	bool m_bExcludeHidden;

public:

	/*!
	** @brief   Initializes a new CGeneralPage object.
	** @remarks -
	*/
	CGeneralPage()
	{
		m_hIcon = NULL;

		m_bDirectory = false;
		m_bRecursive = false;

		m_bExcludeSystem = false;
		m_bExcludeHidden = true;
	}

	/*!
	** @brief   Exchanges and validates dialog data.
	** @param   [in] bSaveAndValidate indicates the direction of the data transfer.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL DoDataExchange(BOOL bSaveAndValidate = FALSE)
	{
		DDX_Check(IDC_RECURSIVE, m_bRecursive, bSaveAndValidate);

		DDX_Check(IDC_EXCLUDE_SYSTEM, m_bExcludeSystem, bSaveAndValidate);
		DDX_Check(IDC_EXCLUDE_HIDDEN, m_bExcludeHidden, bSaveAndValidate);

		if (CGeneralOptions::DoDataExchange(bSaveAndValidate))
		{
			return TRUE;
		}
		return FALSE;
	}

	BEGIN_MSG_MAP(CGeneralPage)
		COMMAND_HANDLER_EX(IDC_RECURSIVE, BN_CLICKED, OnChange)
		COMMAND_HANDLER_EX(IDC_EXCLUDE_SYSTEM, BN_CLICKED, OnChange)
		COMMAND_HANDLER_EX(IDC_EXCLUDE_HIDDEN, BN_CLICKED, OnChange)
		CHAIN_MSG_MAP(CPropertyPageEx<CGeneralPage>)
		CHAIN_MSG_MAP(CGeneralOptions)
		REFLECT_NOTIFICATIONS()
	END_MSG_MAP()

	/*!
	** @brief   Called when no messages are pending on the message queue.
	** @returns TRUE to do further processing, or FALSE otherwise.
	** @remarks -
	*/
	virtual BOOL OnIdle()
	{
		;

		return FALSE;
	}

	/*!
	** @copydoc CPropertyPageEx<CGeneralPage>::OnInitDialog
	*/
	virtual BOOL OnInitDialog(HWND hWndFocus, LPARAM lInitParam)
	{
		CPropertyPageEx<CGeneralPage>::OnInitDialog(hWndFocus, lInitParam);

		CGeneralOptions::OnInitDialog(hWndFocus, lInitParam);

		DDX_Control(IDC_DATETIME_CREATIONTIME,   m_wndDateTimePickers[0], DDX_LOAD);
		DDX_Control(IDC_DATETIME_LASTACCESSTIME, m_wndDateTimePickers[1], DDX_LOAD);
		DDX_Control(IDC_DATETIME_LASTWRITETIME,  m_wndDateTimePickers[2], DDX_LOAD);

		DDX_Icon(IDC_BANNER, m_hIcon, DDX_LOAD);

		ATLVERIFY(DDX_Text(IDC_DISPLAYNAME, m_strDisplayName, 0, DDX_LOAD));
		ATLVERIFY(DDX_Text(IDC_TYPENAME,    m_strTypeName,    0, DDX_LOAD));
		ATLVERIFY(DDX_Text(IDC_LOCATION,    m_strLocation,    0, DDX_LOAD));

		if (m_bDirectory)
		{
			;
		}
		else
		{
			HideDlgCtrl(IDC_RECURSIVE);

			HideDlgCtrl(IDC_EXCLUDE_SYSTEM);
			HideDlgCtrl(IDC_EXCLUDE_HIDDEN);
		}
		ATLVERIFY(DoDataExchange(DDX_LOAD));

		ATLVERIFY(GetPropertySheet().CenterWindow());

		return TRUE;
	}

	/*!
	** @copydoc CPropertyPageEx<CGeneralPage>::OnDestroy
	*/
	virtual void OnDestroy()
	{
		CPropertyPageEx<CGeneralPage>::OnDestroy();

		;
	}
};

/*!
** @brief   The implementation of the Advanced property page.
** @remarks -
*/
class CAdvancedPage :
		public CPropertyPageEx<CAdvancedPage>,
		public CAdvancedOptions<CAdvancedPage>
{
public:

	enum {IDD = IDD_ADVANCED}; // specifies the dialog template resource ID.

public:

	typedef CAdvancedOptions<CAdvancedPage> CAdvancedOptions; ///< A synonym for the template class.
	typedef CAdvancedContext<CAdvancedPage> CAdvancedContext; ///< A synonym for the template class.

	typedef CAdvancedOptions CHandlingOptions; ///< The abstract name of the CAdvancedOptions class.
	typedef CAdvancedContext CHandlingContext; ///< The abstract name of the CAdvancedContext class.

private:

	friend class CAdvancedOptions;
	friend class CDialog;

	CDateTimePickerCtrlExImpl m_wndDateTimePickers[3];

	HICON m_hIcon;

	CString m_strFileSpec;
	CString m_strLocation;

public:

	/*!
	** @brief   Initializes a new CAdvancedPage object.
	** @remarks -
	*/
	CAdvancedPage()
	{
		m_psp.dwFlags |= PSP_PREMATURE;

		m_hIcon = NULL;
	}

	/*!
	** @brief   Exchanges and validates dialog data.
	** @param   [in] bSaveAndValidate indicates the direction of the data transfer.
	** @returns TRUE if successful, or FALSE otherwise.
	** @remarks -
	*/
	BOOL DoDataExchange(BOOL bSaveAndValidate = FALSE)
	{
		ATLVERIFY(DDX_Text(IDC_FILESPEC, m_strFileSpec, 0, bSaveAndValidate));

		if (CAdvancedOptions::DoDataExchange(bSaveAndValidate))
		{
			return TRUE;
		}
		return FALSE;
	}

	BEGIN_MSG_MAP(CAdvancedPage)
		COMMAND_HANDLER_EX(IDC_FILESPEC, EN_CHANGE, OnChange)
		CHAIN_MSG_MAP(CPropertyPageEx<CAdvancedPage>)
		CHAIN_MSG_MAP(CAdvancedOptions)
		REFLECT_NOTIFICATIONS()
	END_MSG_MAP()

	/*!
	** @brief   Called when no messages are pending on the message queue.
	** @returns TRUE to do further processing, or FALSE otherwise.
	** @remarks -
	*/
	virtual BOOL OnIdle()
	{
		;

		return FALSE;
	}

	/*!
	** @copydoc CPropertyPageEx<CAdvancedPage>::OnInitDialog
	*/
	virtual BOOL OnInitDialog(HWND hWndFocus, LPARAM lInitParam)
	{
		CPropertyPageEx<CAdvancedPage>::OnInitDialog(hWndFocus, lInitParam);

		CAdvancedOptions::OnInitDialog(hWndFocus, lInitParam);

		DDX_Control(IDC_DATETIME_CREATIONTIME,   m_wndDateTimePickers[0], DDX_LOAD);
		DDX_Control(IDC_DATETIME_LASTACCESSTIME, m_wndDateTimePickers[1], DDX_LOAD);
		DDX_Control(IDC_DATETIME_LASTWRITETIME,  m_wndDateTimePickers[2], DDX_LOAD);

		DDX_Icon(IDC_BANNER, m_hIcon, DDX_LOAD);

		ATLVERIFY(DDX_Text(IDC_LOCATION, m_strLocation, 0, DDX_LOAD));

		ATLVERIFY(DoDataExchange(DDX_LOAD));

		return TRUE;
	}

	/*!
	** @copydoc CPropertyPageEx<CAdvancedPage>::OnDestroy
	*/
	virtual void OnDestroy()
	{
		CPropertyPageEx<CAdvancedPage>::OnDestroy();

		;
	}
};

/*!
** @brief   The main dialog box.
** @remarks -
*/
class CDialog : public CPropertySheetEx
{
public:

	CGeneralPage  m_general;  ///< The General property page.
	CAdvancedPage m_advanced; ///< The Advanced property page.

private:

	typedef CGeneralPage::CGeneralOptions CGeneralOptions;
	typedef CGeneralPage::CGeneralContext CGeneralContext;

	typedef CAdvancedPage::CAdvancedOptions CAdvancedOptions;
	typedef CAdvancedPage::CAdvancedContext CAdvancedContext;

	CFile m_file;
	CIcon m_icon;

public:

	/*!
	** @brief   Constructs a new CDialog object that is bound to the specified file.
	** @param   [in] path the absolute path to the target file or directory.
	** @remarks -
	*/
	explicit CDialog(const CPath& path)
	{
		m_file = path;

		m_general.m_bDirectory = m_file.IsDirectory() ? true : false;

		SHFILEINFO sfi = {};

		ATLVERIFY(m_file.GetFileInfo(sfi, SHGFI_ICON | SHGFI_DISPLAYNAME | SHGFI_TYPENAME));

		m_icon = sfi.hIcon;
		ATLASSERT(m_icon);
		m_advanced.m_hIcon = m_general.m_hIcon = m_icon;

		m_general.m_strDisplayName = sfi.szDisplayName;
		ATLASSERT(!m_general.m_strDisplayName.IsEmpty());
		m_general.m_strTypeName = sfi.szTypeName;
		ATLASSERT(!m_general.m_strTypeName.IsEmpty());
		m_general.m_strLocation = CString(m_file.GetParent());
		ATLASSERT(!m_general.m_strLocation.IsEmpty());
		m_advanced.m_strLocation = CString(m_file);
		ATLASSERT(!m_advanced.m_strLocation.IsEmpty());

		CFileTimeEx ftCurrentTime = CFileTimeEx::GetCurrentTime().ZeroMillisecond();

		m_advanced.m_ftCurrentTime = m_general.m_ftCurrentTime = ftCurrentTime;

		CFileTimeEx ftCreationTime;
		CFileTimeEx ftLastAccessTime;
		CFileTimeEx ftLastWriteTime;

		if (m_file.GetFileTime(&ftCreationTime, &ftLastAccessTime, &ftLastWriteTime))
		{
			m_advanced.m_ftCreationTime   = m_general.m_ftCreationTime   = ftCreationTime;
			m_advanced.m_ftLastAccessTime = m_general.m_ftLastAccessTime = ftLastAccessTime;
			m_advanced.m_ftLastWriteTime  = m_general.m_ftLastWriteTime  = ftLastWriteTime;
		}
		else // fallback
		{
			m_advanced.m_ftCreationTime   = m_general.m_ftCreationTime   = ftCurrentTime;
			m_advanced.m_ftLastAccessTime = m_general.m_ftLastAccessTime = ftCurrentTime;
			m_advanced.m_ftLastWriteTime  = m_general.m_ftLastWriteTime  = ftCurrentTime;
		}
		if (m_general.m_bDirectory)
		{
			ATLVERIFY(AddPage(m_general));
			ATLVERIFY(AddPage(m_advanced));
		}
		else
		{
			ATLVERIFY(AddPage(m_general));
		}
	}

	/*!
	** @copydoc CPropertySheetEx::OnApply
	*/
	virtual int OnApply()
	{
		int iResult = PSNRET_INVALID_NOCHANGEPAGE;

		CFileTimeEx ftCurrentTime = CFileTimeEx::GetCurrentTime().ZeroMillisecond();

		m_advanced.m_ftCurrentTime = m_general.m_ftCurrentTime = ftCurrentTime;

		CFileTimeEx ftCreationTime;
		CFileTimeEx ftLastAccessTime;
		CFileTimeEx ftLastWriteTime;

		if (m_file.IsDirectory())
		{
			CWaitCursor cursor;

			LPCTSTR pszFileSpec = m_advanced.m_strFileSpec.IsEmpty() ? NULL : m_advanced.m_strFileSpec;

			if (Traverse(m_file, pszFileSpec, m_general.m_bRecursive))
			{
				if (!m_file.GetFileTime(&ftCreationTime, &ftLastAccessTime, &ftLastWriteTime))
				{
					goto OnError;
				}
				m_general.m_ftCreationTime   = ftCreationTime;
				m_general.m_ftLastAccessTime = ftLastAccessTime;
				m_general.m_ftLastWriteTime  = ftLastWriteTime;

				ATLVERIFY(m_general.DoDataExchange(DDX_LOAD));

				iResult = PSNRET_NOERROR;
			}
		}
		else
		{
			CGeneralContext general = m_general;

			if (m_file.PrepareForWrite(m_hWnd, SHPPFW_ASKDIRCREATE | SHPPFW_IGNOREFILENAME))
			{
				m_file.EnsureFileExists();

				if (!m_file.GetFileTime(&ftCreationTime, &ftLastAccessTime, &ftLastWriteTime))
				{
					goto OnError;
				}
				if (!general.HasNoOperations())
				{
					general.ModifyFileTime(ftCreationTime, ftLastAccessTime, ftLastWriteTime);

					if (!m_file.SetFileTime(&ftCreationTime, &ftLastAccessTime, &ftLastWriteTime)
					 || !m_file.GetFileTime(&ftCreationTime, &ftLastAccessTime, &ftLastWriteTime))
					{
						goto OnError;
					}
				}
				m_general.m_ftCreationTime   = ftCreationTime;
				m_general.m_ftLastAccessTime = ftLastAccessTime;
				m_general.m_ftLastWriteTime  = ftLastWriteTime;

				ATLVERIFY(m_general.DoDataExchange(DDX_LOAD));

				iResult = PSNRET_NOERROR;
			}
		}
		return iResult;

	OnError:

		HRESULT hResult = AtlHresultFromLastError();
		Alert(hResult);
		return iResult;
	}

	/*!
	** @brief   Traverses a directory tree to process all the files and subdirectories.
	** @param   [in] pszDir the path to a directory to traverse.
	** @param   [in] pszFileSpec a semicolon-delimited list of file name patterns, or NULL for none.
	** @param   [in] bRecursive specifies whether to traverse recursively or not.
	** @returns TRUE to continue the traverse, or FALSE otherwise.
	** @remarks -
	*/
	BOOL Traverse(LPCTSTR pszDir, LPCTSTR pszFileSpec = NULL, BOOL bRecursive = TRUE)
	{
		CGeneralContext  general  = m_general;
		CAdvancedContext advanced = m_advanced;

		if (general.HasNoOperations() && advanced.HasNoOperations()) // entirely no operations
		{
			return TRUE;
		}
		CFindFileEx find;

		CFile file;

		CFileTimeEx ftCreationTime;
		CFileTimeEx ftLastAccessTime;
		CFileTimeEx ftLastWriteTime;

		bool bEmpty = true;

		for (BOOL bFound = find.FindFile(pszDir, TEXT("*.*")); bFound; bFound = find.FindNextFile())
		{
			if ((m_general.m_bExcludeSystem && find.IsSystem())
			 || (m_general.m_bExcludeHidden && find.IsHidden()))
			{
				continue;
			}
			if (find.IsDirectory())
			{
				if (find.IsDots()) // represents the current directory or its parent directory
				{
					continue;
				}
				file = LPCTSTR(find.GetFilePath());

				if (bRecursive)
				{
					if (Traverse(file, pszFileSpec))
					{
						if (!file.GetFileTime(&ftCreationTime, &ftLastAccessTime, &ftLastWriteTime))
						{
							goto OnError;
						}
					}
					else // failed to traverse the directory tree
					{
						return FALSE;
					}
				}
				else
				{
					ATLTRACE(TEXT("-%s\n"), LPCTSTR(file));

					ftCreationTime   = find.m_fd.ftCreationTime;
					ftLastAccessTime = find.m_fd.ftLastAccessTime;
					ftLastWriteTime  = find.m_fd.ftLastWriteTime;
				}
			}
			else
			{
				file = LPCTSTR(find.GetFilePath());

				ATLTRACE(TEXT("-%s\n"), LPCTSTR(file));

				ftCreationTime   = find.m_fd.ftCreationTime;
				ftLastAccessTime = find.m_fd.ftLastAccessTime;
				ftLastWriteTime  = find.m_fd.ftLastWriteTime;

				if (!pszFileSpec || find.MatchesSpec(pszFileSpec, PMSF_MULTIPLE))
				{
					if (!advanced.HasNoOperations())
					{
						advanced.ModifyFileTime(ftCreationTime, ftLastAccessTime, ftLastWriteTime);

						if (!file.SetFileTime(&ftCreationTime, &ftLastAccessTime, &ftLastWriteTime)
						 || !file.GetFileTime(&ftCreationTime, &ftLastAccessTime, &ftLastWriteTime))
						{
							goto OnError;
						}
					}
				}
			}
			if (general.m_ftEarliestCreationTime   > ftCreationTime)
				general.m_ftEarliestCreationTime   = ftCreationTime;
			if (general.m_ftEarliestLastAccessTime > ftLastAccessTime)
				general.m_ftEarliestLastAccessTime = ftLastAccessTime;
			if (general.m_ftEarliestLastWriteTime  > ftLastWriteTime)
				general.m_ftEarliestLastWriteTime  = ftLastWriteTime;

			if (general.m_ftLatestCreationTime     < ftCreationTime)
				general.m_ftLatestCreationTime     = ftCreationTime;
			if (general.m_ftLatestLastAccessTime   < ftLastAccessTime)
				general.m_ftLatestLastAccessTime   = ftLastAccessTime;
			if (general.m_ftLatestLastWriteTime    < ftLastWriteTime)
				general.m_ftLatestLastWriteTime    = ftLastWriteTime;

			bEmpty = false;
		}
		file = pszDir;

		ATLTRACE(TEXT("+%s\n"), LPCTSTR(file));

		if (!file.GetFileTime(&ftCreationTime, &ftLastAccessTime, &ftLastWriteTime))
		{
			goto OnError;
		}
		if (bEmpty)
		{
			general.m_ftEarliestCreationTime   = ftCreationTime;
			general.m_ftEarliestLastAccessTime = ftLastAccessTime;
			general.m_ftEarliestLastWriteTime  = ftLastWriteTime;

			general.m_ftLatestCreationTime     = ftCreationTime;
			general.m_ftLatestLastAccessTime   = ftLastAccessTime;
			general.m_ftLatestLastWriteTime    = ftLastWriteTime;
		}
		if (!general.HasNoOperations())
		{
			general.ModifyFileTime(ftCreationTime, ftLastAccessTime, ftLastWriteTime);

			if (!file.SetFileTime(&ftCreationTime, &ftLastAccessTime, &ftLastWriteTime))
			{
				goto OnError;
			}
		}
		return TRUE;

	OnError:

		HRESULT hResult = AtlHresultFromLastError();
		Alert(hResult, file);
		return FALSE;
	}
};
